// Parser.h a parser object in the Parser namespace, cannot be used directly
// in straight C.

/** \file Parser.h

 This file uses C++ namespaces to make the global object pointer and parsing
 routine names unique.

 To mix with other code, prefix the function name as in Parser::FuncName()
 or put 'using namespace Parser' inside the function body, or if in a 
 dedicated file where no name conflicts are likely, set the namespace for 
 the whole file with 'using namespace Parser;'.

 Note: if you work directly with obj.op make sure to call obj_needs(n)
 when adding to it to assure the buffer can hold the new data.

 Don't forget to do
 - namespace Parser { ... }
 or
 - using namespace Parser;
 in your code in order to access these functions. 

 Also, since the global object is renamed 'obj' in the header for this file
 that header should be loaded last in a list of includes to prevent perplexing 
 syntax error messages that may be caused by this usurpation of a parameter list 
 name or something of the sort.

*/


#ifndef Parser_h
#define Parser_h

#include <unistd.h> // NULL

#ifndef __cplusplus
#warning %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#warning This file requires C++ namespaces
#warning %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#endif

// '#define NO_PARSER_WARN 1' to silence this warning.
#if !NO_PARSER_WARN
#warning INCLUDE THIS FILE LAST to avoid name conflicts due to 'obj' macro
#endif

namespace Parser
{

/////////////////////////////////////////////
// data structures

typedef struct {
  bool state;       /// current match state
  const char* ip;   /// read-only input pointer
  char* op;         /// output pointer
  const char* ibuf; /// read-only input buffer
  char* obuf;       /// output buffer base
  int iblen;        /// input buffer length
  int oblen;        /// output buffer length
  const char* iblimit;  /// end of input buffer
  char* oblimit;    /// threshold for output buffer reallocation
  bool dynalloc;    /// input buffer can be freed
}PARSER_OBJ;

typedef struct {
  const char* ip;  /// saved input pointer
  char* op;        /// saved output pointer
}PPTRS_OBJ;

/// trick to allow dotted field addressing through the global pointer,
/// e,g, 'obj.fieldName' instead of 'g_parser_obj->fieldName'.  This 
/// is also a convenient watch point for working out parsing algorithms.
#define obj (*Parser::g_parser_obj)
extern PARSER_OBJ* g_parser_obj;

////////////////////////////////////////////////////////////
// init function for user supplied input buffer

/// returns previous obj
PARSER_OBJ* parser_init(
        PARSER_OBJ* This,         /// addr of object to use
        const char* inbuffer,     /// initialized text to parse
        int inlen,                /// exact length including terminating null
        bool dynalloc             /// whether or not we can free inbuffer when done
        );


/// frees memory and restores previous global object if not NULL
void parser_free(PARSER_OBJ* This, PARSER_OBJ* Prev = NULL);

/// restores a previous global object pointer
PARSER_OBJ* parser_set_current(PARSER_OBJ* This);

/// internal function for SAVE_PTRS()
PPTRS_OBJ* pvt_parser_save_ptrs(PPTRS_OBJ* p);
/// internal function for RESTORE_PTRS()
void pvt_parser_restore_ptrs(PPTRS_OBJ* p);

/// creates a save struct in current scope and saves ip/op pointers
#define SAVE_PTRS() \
  PPTRS_OBJ pvt_ptrs; pvt_ptrs = *pvt_parser_save_ptrs(&pvt_ptrs)

/// uses a save struct in current scope to restore ip/op pointers
#define RESTORE_PTRS() \
  pvt_parser_restore_ptrs(&pvt_ptrs)
  

////////////////////////////////////////////////////////
// built-in utilities acting on ip and/or op - prefix 'obj'
// is -- test IF all match (changes: none)
// copy -- test and copy to op IF all match (changes; ip, op)
// skip -- test and skip IF all match (changes: ip)

/// ip is string s, else false
bool obj_is(const char* s);

/// copy s if at ip, else false
bool obj_copy(const char* s);

/// skip s if at ip, else false
bool obj_skip(const char* s);

/// true if there are n chars or more left to go
bool obj_is_nchars(int n);

/// copy n chars from ip to op
bool obj_copy_nchars(int n);

/// skip n chars or to end of buffer if n > what's left at ip.
bool obj_skip_nchars(int n);
  
//////////////////////////////////////////////////////
// special functions, always return true

/// Dynamic reallocation check, always returns true.  Allocates
/// bytes as needed for output.
bool obj_needs(int numBytes);

/// put text in output buffer, always returns true
bool obj_put(const char* text);

/// puts a single character in output buffer, always returns true
bool obj_put_char (int c);

///////////////////////////////////////////////////////////////
// built-in std parser funcs (no prefix)
// type bool func(void) for possible use by name in grammars)

/// end of input
bool obj_is_eoi();   

/// not end of input
bool obj_not_eoi();  

//////////////////////////////////////////////////////////////

/// is whitespace at ip, else false
bool obj_is_ws();

/// copy whitespace at ip, else false
bool obj_copy_ws();

/// skip whitespace at ip, else false
bool obj_skip_ws();

//////////////////////////////////////////////////////////////

/// is text to end of line, not including newline, alias for not_eoi()
bool obj_is_linetext();

/// copy text to end of line, not including newline
bool obj_copy_linetext();

/// skip text to end of line, not including newline
bool obj_skip_linetext();

//////////////////////////////////////////////////////////////

/// is a newline at ip, else false
bool obj_is_newline();

/// copy a newline at ip, else false
bool obj_copy_newline();

/// skip a newline at ip, else false
bool obj_skip_newline();

//////////////////////////////////////////////////////////////

// functions to handle delimited text (handles escaped chars, 
// such as single and double quotes within a string), up to 
// but not including delimiter.  Delimiter can be any string.

/// delim exists in ip, else false
bool obj_is_to(const char* delim);

/// copy to delim in ip, else false
bool obj_copy_to(const char* delim);

/// skip to delim in ip, else false
bool obj_skip_to(const char* delim);

//////////////////////////////////////////////////////////////

// functions working on text at output returning 
// an (optional) temp string that will be recycled if not used
// immediately (up to 32 strings before recycling).  These
// always return true if no errors.

/// cut from start to end at op, true if no errors
bool obj_cut_op(char** ret, char* start, char* end);

/// copy from start to end at op, true if no errors
bool obj_copy_op(char** ret, char* start, char* end);

/// delete from start to end at op, true if no errors
bool obj_delete_op(char* start, char* end);

/// past string \b from into op at \b to
bool obj_paste_op(char* to, char* from);

// convenience funcs

/// set's pointer to string in input or null if none, returns true if found
/// \Note: scans forward
bool obj_find_ip(const char** ret, const char* s);

/// set's pointer to string in output or null if none, returns true 
/// \Note: scans backward. Used to remove '/name' + '/..' pairs in 
/// file paths, etc.
bool obj_find_op_back(char** ret, const char* s);

/// The safe way to return obj.obuf, makes sure the string is
/// terminated.  returns a pointer to obj.obuf, which must be 
/// copied before the the object mem is freed.
char* obj_get_result();

} // namespace Parser

#endif // Parser_h

