// dirlist.h list of type/name pairs

#ifndef LQ_dirlist_h
#define LQ_dirlist_h

#include <string.h> // strdup()
#include <malloc.h> // free()

#include <LQ/objlist.h>
#include <LQ/cdecls.h>
#include <LQ/types.h> // bool for C if needed

/** \file dirlist.h
 * Reads and stores stripped down directory information for files consisting
 * of only the name and type of item.  The types can be identified using 
 * S_ISFIFO, S_ISCHR, etc., per sys/stat.h declarations.
 */

LQ_BEGIN_CDECLS;


typedef struct
{
  const char* name;
  int type;
}DIRLIST_ITEM;

typedef DIRLIST_ITEM** DIRLIST;

enum
{
  DL_SORT_NONE = 0,
  DL_SORT_ALPHA = 1,
  DL_SORT_CASE = 1 << 1,
  DL_SORT_DIR = 1 << 2,
  DL_NO_HIDDEN = 1 << 3,
  DL_NO_IRREG = 1 << 4,
  DL_SORT_DEFAULT_PLUS_HIDDEN = DL_SORT_DIR | DL_SORT_ALPHA,
  DL_SORT_DEFAULT = DL_SORT_DIR | DL_SORT_ALPHA | DL_NO_HIDDEN
};

// Macros for type values defined in <sys/stat.h>
//    UNKNOWN = 0,
//    S_ISFIFO / S_IFIFO = 1,
//    S_ISCHR / S_IFCHR = 2,
//    S_ISDIR / S_IFDIR = 4,
//    S_ISBLK / S_IFBLK = 6,
//    S_ISREG / S_IFREG = 8,
//    S_ISLNK / S_IFLNK = 10,
//    S_ISSOCK / S_IFSOCK = 12,

    inline void dirlist_remove(DIRLIST* plist, int idx)
      {objlist_remove((OBJLIST*)plist, idx);}
    inline int dirlist_last(DIRLIST* plist)
      {return objlist_last((OBJLIST*)plist);}
    
    DIRLIST dirlist_new();
    
    /// inits and reads dir using DL_* flags for sorting and showing
    int dirlist_read(DIRLIST* plist, const char* dirname, const char* pattern /* e.g., "*"*/, int sort_method /* e.g., -1 = default */);
    
    /// The proper way to free a list.
    inline void dirlist_delete(DIRLIST* plist)
      {objlist_delete((OBJLIST*)plist);}
    
    /// clears the list for reuse.
    inline void dirlist_clear(DIRLIST* plist)
      {objlist_clear((OBJLIST*)plist);}
    
    /// gets the number of items in the list
    inline int dirlist_getCount(DIRLIST* plist) 
      {return objlist_getCount((OBJLIST*)plist);}

    /// adds a new item to the list
    void dirlist_append(DIRLIST* plist, DIRLIST_ITEM* o);
      
    /// creates and adds a new item consisting of name and type.
    void dirlist_addNew(DIRLIST* plist, const char* name, int type);
    
    /// inerts and item at idx
    inline void dirlist_insert(DIRLIST* plist, int idx, DIRLIST_ITEM* o)
      {objlist_insert((OBJLIST*)plist, (OBJLIST_ITEM*)o, idx);}
      
    /// returns index of the object containing the name in the list, returns -1 if not 
    /// found.  The parameter start = where to start the search (0 = at the beginning 
    /// of the list).
    int dirlist_find(DIRLIST* plist, const char* name, int opt_start);  
    
    /// returns index of first item in the array, always 0 if there are any 
    /// items, else -1.
    inline int dirlist_first(DIRLIST* plist) {return objlist_first((OBJLIST*)plist);}
    
    /// returns the DIRLIST_ITEM* at idx else null.
    DIRLIST_ITEM* dirlist_get(DIRLIST* plist, int idx);

    /// returns the name of the item at idx else null.
    const char* dirlist_getName(DIRLIST* plist, int idx);
    
    /// returns linux standard type id from item at idx, 8 = regular file, 4 = dir,
    /// etc.
    int dirlist_getType(DIRLIST* plist, int idx);
    
    /// replaces or inits one list with another
    void dirlist_set(DIRLIST* plist, int idx, DIRLIST_ITEM* dirlist_new_obj); 
    
    /// checks item at idx and returns true if it's a dir
    bool dirlist_isDir(DIRLIST* plist, int idx);

    /// checks item at idx and returns true if it's a regular file.
    bool dirlist_isReg(DIRLIST* plist, int idx);
    
LQ_END_CDECLS;

#endif // LQ_dirlist_h
