// objlist.h

#ifndef objlist_h
#define objlist_h

/** \file objlist.h
 * Note that the OBJLIST item is type void, as a char is an element
 * in a string.  We never use the item directly but alsoways the 
 * OBJLIST_ITEM*, which is the analogy of a char*.
 * 
 * See an example of a simple list in slist.c and slist.h and a more
 * complex example (structs with destructors) in dirlist.h and dirlist.c
*/

#include <LQ/cdecls.h>

LQ_BEGIN_CDECLS;

typedef void OBJLIST_ITEM;
typedef OBJLIST_ITEM** OBJLIST;

/// destructor function type called to free items in a list.  Used per 
/// item for deletion of individual items or the whole array.
typedef void (*objlist_dtor_t)(OBJLIST_ITEM* item);

/// Sets item destructor field for the list so objlist can handle
/// freeing items and arrays.
void objlist_setDestructor(OBJLIST* plist, objlist_dtor_t func);


/// returns index of first item (always 0)
int objlist_first(OBJLIST* plist);

/// returns index of last item (count -1), returns -1 if none.
int objlist_last(OBJLIST* plist);

/// returns item at index
OBJLIST_ITEM* objlist_get(OBJLIST* plist, int index);

/// USE WITH CAUTION: sets item at index, must be same type as previous
void _objlist_set(OBJLIST* plist, int index, OBJLIST_ITEM* item);

/// returns a pointer to the first object in a new list.
OBJLIST_ITEM** objlist_new();

/// removes item and deletes the base
void objlist_delete(OBJLIST* plist);

/// removes items, does not resize or delete the base
void objlist_clear(OBJLIST* plist);

/// returns number of items in the list    
int objlist_count(OBJLIST list);

/// returns number of items in the list    
int objlist_getCount(OBJLIST* list);

/// Adds a new items at the end of the list. The new item must 
/// have it's own memory.
void objlist_append(OBJLIST* plist, OBJLIST_ITEM* o);

/// removes (and frees) item at idx, if any. Does not resize the list.
void objlist_remove(OBJLIST* plist, int idx);

// Removed this.  Can't do without knowing size of object (ctor)
// adds list list2 to list1, does not delete list that's added
// void objlist_addList(OBJLIST* plist1, OBJLIST* plist2);

/** 
  Inserts new item at idx, or if out of range, at end or at beginning.
  User supplies error checking if out-of-range subscripts need to be
  handled.  The new item must have it's own memory.
*/
void objlist_insert(OBJLIST* plist, OBJLIST_ITEM* o, int idx);

/// removes to end from idx, inclusive
void objlist_trimRight(OBJLIST* plist, int idx);

/// removes from beginning to idx, inclusive
void objlist_trimLeft(OBJLIST* plist, int idx);

LQ_END_CDECLS;

#endif // objlist_h

