// parser.cpp a parser object in the Parser namespace

#include <malloc.h>
#include <string.h>

#define NO_PARSER_WARN 1
#include <LQ/Parser.h> // load last in list

namespace Parser
{
  
#define OBUF_INIT_SIZE 4096
#define OBUF_PADDING 256

// required global object
PARSER_OBJ* g_parser_obj;

////////////////////////////////////////////////////
// built-in functions

static void _parser_init() {
  obj.state = true;
  obj.ip = obj.ibuf;
  obj.op = obj.obuf;
  *obj.op = 0;
}

// Returns previous obj
PARSER_OBJ* parser_init(
  PARSER_OBJ* This, 
  const char* inbuffer, 
  int inlen, 
  bool dynalloc
)
{
  PARSER_OBJ* prev = g_parser_obj;
  g_parser_obj = This;
  obj.iblen = inlen;
  obj.ibuf = inbuffer;
  obj.iblimit = obj.ibuf + obj.iblen;
  obj.dynalloc = dynalloc;
  
  obj.oblen = OBUF_INIT_SIZE;
  obj.obuf = (char*)malloc(OBUF_INIT_SIZE);
  obj.oblimit = obj.obuf + OBUF_INIT_SIZE - OBUF_PADDING;
  _parser_init();
  return prev;
}

void parser_free(PARSER_OBJ* This, PARSER_OBJ* Prev)
{
  if(This->obuf) free(This->obuf);
  if(This->dynalloc)
    free((void*)This->ibuf);
  memset(This, 0, sizeof(*This));
  g_parser_obj = Prev;
}

// restores a previous global object pointer
PARSER_OBJ* parser_set_current(PARSER_OBJ* This)
{
  PARSER_OBJ* Prev = g_parser_obj;
  g_parser_obj = This;
  return Prev;
}


bool obj_needs(int numBytes)
{
  if((numBytes + obj.op) > obj.oblimit)
  {
    int newsize = obj.oblen * 2;
    int op_offset = obj.op - obj.obuf;
    obj.obuf = (char*)realloc(obj.obuf, newsize);
    obj.oblen = newsize;
    obj.oblimit = obj.obuf + newsize - OBUF_PADDING;
    obj.op = obj.obuf + op_offset;
  }
  return obj.state = true;
}

bool obj_put(const char* text)
{
  int slen = strlen(text);
  obj_needs(slen);
  strcpy(obj.op, text);
  obj.op += slen;
  return obj.state = true;
}

bool obj_put_char (int c)
{
  obj_needs(1);
  *obj.op++ = c;
  return obj.state;
}

bool obj_is_eoi() {
  return obj.state = (obj.ip >= obj.iblimit) ? true : false;
}

bool obj_not_eoi() { 
  return obj.state = !obj_is_eoi();
}

PPTRS_OBJ* pvt_parser_save_ptrs(PPTRS_OBJ* p) {
  p->ip = obj.ip;
  p->op = obj.op;
  return p;
}

void pvt_parser_restore_ptrs(PPTRS_OBJ* p) {
  obj.ip = p->ip;
  obj.op = p->op;
}

// true if string at ip = s
bool obj_is(const char* s) {
  obj.state = true;
  int slen = strlen(s);
  if(!obj_is_nchars(slen)) // check ip is long enough to be it
    return obj.state = false;

  if(memcmp((void*)obj.ip, (void*)s, slen) != 0)
    obj.state = false;
  return obj.state;
}

// copy if ip = s
bool obj_copy(const char* s) {
  if(!obj_is(s))    
    return obj.state;
                    
  int slen = strlen(s);
  obj_needs(slen);
  
  strcpy(obj.op, s);
  obj.op += slen;
  obj.ip += slen;
  return obj.state = true;
}


// skip if ip = s
bool obj_skip(const char* s) {
  if(!obj_is(s))
    return obj.state;
  int slen = strlen(s);
  obj.ip += slen;
  return obj.state = true;
}


bool obj_copy_nchars(int n) {
  if(!obj_is_nchars(n))
    return obj.state;
  
  obj_needs(n);
  while(n--)
    *obj.op++ = *obj.ip++;
  *obj.op = 0;
  return obj.state = true;
}

// true if input has n or more chars left
bool obj_is_nchars(int n) {
  return obj.state = ((obj.ip + n) > (obj.iblimit)) ? false : true;
}

bool obj_skip_nchars(int n)
{
  obj.ip += n;
  if(obj.ip < (obj.iblimit))
    return obj.state = true;
  return obj.state = false;
}

// whitespaces
bool obj_is_ws() {
  if(!strchr(" \t\r", *obj.ip))
return obj.state = false;
return obj.state = true;
}

bool obj_skip_ws() {
  if(!strchr(" \t\r", *obj.ip))
return obj.state = false;
obj.ip++;
return obj.state = true;
}

bool obj_copy_ws() {
  if(!obj_is_ws())
    return obj.state;
  obj_needs(1);
  *obj.op++ = *obj.ip++;
  return obj.state = true;
}

// text to end of line, not including newline
bool obj_is_linetext(){return obj_not_eoi();} 

bool obj_skip_linetext()
{
  if(obj_is_eoi())
    return obj.state = false;
  // do all at once
  const char* eol = strchr(obj.ip, '\n');
  if(!eol)
    eol = obj.ip + strlen(obj.ip);
  obj_skip_nchars(eol - obj.ip);
  return obj.state; // if nchars
}

bool obj_copy_linetext()
{
  if(obj_is_eoi())
    return obj.state = false;
  // do all at once
    const char* eol = strchr(obj.ip, '\n');
    if(!eol)
      eol = obj.ip + strlen(obj.ip);
    obj_copy_nchars(eol - obj.ip);
    return obj.state; // if nchars
}

bool obj_is_newline() { return obj_is("\n"); }
bool obj_skip_newline() { return obj_skip("\n"); }
bool obj_copy_newline() { return obj_copy("\n"); }

// mid-level functions to handle delimited text (ignores escaped 
// chars, such as single and double quotes within a string), up 
// to but not including delimiter.  Delimiter can be any string.

// set's pointer to string in input or null if none, returns true if found
// scans forward
bool obj_find_ip(const char** ret, const char* s) {
  *ret = false;
  int slen = strlen(s);
  const char* end = obj.ip + obj.iblen - slen - 1;
  const char* p = obj.ip;
  for( ; p < end; p++) {
    if(memcmp((void*)p, (void*)s, slen) == 0) {
      if(p == obj.ibuf) {
        *ret = p;
        return obj.state = true;
      }
      else if(p[-1] != '\\') {
        *ret = p;
        return obj.state = true;
      }
    }
  }
  return obj.state = false;
}

// set's pointer to string in output or null if none, returns true 
// if found scans backward
bool obj_find_op_back(char** ret, const char* s) {
  *ret = 0;
  int slen = strlen(s);
  char* p = obj.op - slen;
  char* end = obj.obuf - 1;
  for(; p > end; p--) {
    if(memcmp((void*)p, (void*)s, slen) == 0) {
      *ret = p;
      return obj.state = true;
    }
  }
  return obj.state = false;
}

// looks backward in op for delim and if found trims output to 
// that point, else returns false.
bool obj_delete_from(char* delim) {
  char* from;
  obj_find_op_back(&from, delim);
  if(!obj.state) return false;
  obj.op = from;
  return obj.state = true;
}

// true if exists in input
bool obj_is_to(const char* delim) { 
  const char* dummy;
  if(obj_find_ip(&dummy, delim))
    return obj.state = true;
  return obj.state = false;
}

bool obj_skip_to(const char* delim) {
  const char* to;
  obj_find_ip(&to, delim);
  if(!obj.state) return false;
  obj.ip = to;
  return obj.state; // true
}

bool obj_copy_to(const char* delim) {
  const char* to;
  obj_find_ip(&to, delim);
  if(!obj.state) return false;

  int slen = to - obj.ip;
  obj_needs(slen);
  memcpy(obj.op, obj.ip, slen);
  obj.op += slen;
  *obj.op = 0; // terminate string
  obj.ip = to;
  return obj.state = true;
}


//////////////////////////////////////////////////////////
// High level functions working on already compiled output
// return a temp string that will be recycled if not used
// immediately (up to 32 strings before recycling).

#define MAX_TMPBUF 32
static char* tmpbuf[MAX_TMPBUF];
static int tmpbufptr;

#define TMPBUF_CURRENT \
tmpbuf[tmpbufptr]

void tmpbuf_next() {
  tmpbufptr++;
  if(tmpbufptr >= MAX_TMPBUF)
    tmpbufptr = 0;
}


bool obj_cut_op(char** ret, char* start, char* end) {
  if(ret)
    obj_copy_op(ret, start, end);
  obj_delete_op(start, end);
  return obj.state = true;
}

bool obj_copy_op(char** ret, char* start, char* end) {
  if(!ret || ! start || !end)
    return obj.state = false;
  
  int slen = end - start;
  if(TMPBUF_CURRENT)
    TMPBUF_CURRENT = (char*)realloc(TMPBUF_CURRENT, slen + 1);
  else
    TMPBUF_CURRENT = (char*)malloc(slen + 1);
  memcpy(TMPBUF_CURRENT, start, slen);
  TMPBUF_CURRENT[slen] = 0; // terminate string
  tmpbuf_next();
  return obj.state = true;
}

bool obj_delete_op(char* start, char* end)
{
  int len = strlen(end);
  memcpy((void*)start, (void*)end, len + 1); // include terminator
  obj.op = start + len;
  return obj.state = true;
}

bool obj_paste_op(char* to, char* from)
{
  int slen = strlen(from);
  // insert len bytes and copy without terminator
  obj_needs(slen + 1);
  memmove(to + slen, to, strlen(to));
  memcpy(to, from, slen);
  return obj.state = true;
}

char* obj_get_result()
{
  *obj.op = 0;
  return obj.obuf;
}

} // namespace Parser

