#include <LQ/dirlist.h>
#include <dirent.h>
#include <stdlib.h>
#include <LQ/types.h>
#include <LQ/filename.h>


int casealphasort(const struct dirent **a, const struct dirent **b)
{
  return strcasecmp(a[0]->d_name, b[0]->d_name);
}


// destructor for the dirlist, needs to free the strdup-ed name field.
void dirlist_dtor(void* item)
{
  DIRLIST_ITEM* obj = (DIRLIST_ITEM*)item;
  if(obj->name)
    free((void*)obj->name);
  obj->name = 0;
}

DIRLIST dirlist_new()
{
  OBJLIST o = objlist_new();
  objlist_setDestructor(&o, dirlist_dtor);
  return (DIRLIST)o;
}


// returns pointer to the item
DIRLIST_ITEM* dirlist_get(DIRLIST* plist, int idx) 
{
  return (DIRLIST_ITEM*) objlist_get((OBJLIST*)plist, idx);
}


const char* dirlist_getName(DIRLIST* plist, int idx) // return name
{
  return dirlist_get(plist, idx)->name;
}


int dirlist_getType(DIRLIST* plist, int idx) // return name
{
  return dirlist_get(plist, idx)->type;
}


void obs_dirlist_dtor(DIRLIST*** plist, void** po)
{
  free(*po);
  *po = 0;
}

static DIRLIST_ITEM* dirlist_newItem(const char* name, int type)
{
  DIRLIST_ITEM* p = (DIRLIST_ITEM*)malloc(sizeof(DIRLIST_ITEM));
  p->name = strdup(name);
  p->type = type;
  return p;
}


// -1 if not found, 
int dirlist_find(DIRLIST* plist, const char* name, int opt_start /* = 0 */ )  
{
  DIRLIST_ITEM* o;
  for(int i = dirlist_first(plist); i < dirlist_getCount(plist); i++)
  {
    o = (*plist)[i];
    if(strcmp(name, o->name) == 0)
      return i;
  }
  return -1;  // not found
}

void dirlist_addNew(DIRLIST* plist, const char* name, int type)
{
  DIRLIST_ITEM* o = dirlist_newItem(name, type);
  dirlist_append(plist, o);
}

void dirlist_append(DIRLIST* plist, DIRLIST_ITEM* o)
{
  objlist_append((OBJLIST*)plist, (OBJLIST_ITEM*)o);  
}


// returns number of entries
int dirlist_read(DIRLIST* plist, const char* a_dirname, const char* pattern /* = "*" */, int sort_method /* = -1 */ )
{
  int n_files;
  dirent* dir;            // for processing below, not entirely necessary
  dirent** entries = 0;   // this does the work, is malloced and must be freed.
  char* dirname = (char*)malloc(strlen(a_dirname) + 256);
  if(a_dirname[0] == '~')
    sprintf(dirname, "%s%s", getenv("HOME"), a_dirname+1);
  else
    sprintf(dirname, "%s", a_dirname);

  // don't filter dirs
  const char* dir_pattern = "*";
  dirlist_clear(plist);
  
  if(sort_method == -1) // sort dirs first, alpha, case insensitive, no hidden
    sort_method = DL_SORT_DEFAULT;

  // first pass sort alpha
  if(sort_method & DL_SORT_ALPHA)
  {
    if(sort_method & DL_SORT_CASE) // case sensitive
      n_files = filename_dirList(dirname, &entries, casealphasort);
    else // case insensitive
      n_files = filename_dirList(dirname, &entries, alphasort);
  }
  else // unsorted
    n_files = filename_dirList(dirname, &entries, 0);
    
  // second pass, turn to DIRLIST only if pattern matches
  // and sort dirs and files, etc.
  if(sort_method & DL_SORT_DIR)
  {
    for(int i = 0; i < n_files; i++)
    {
      dir = entries[i];
      if(dir->d_type == DT_DIR) // dirs first
      {
        if(filename_matchExisting(dir->d_name, dir_pattern))
          dirlist_addNew(plist, dir->d_name, dir->d_type);
      }
    }
    
    for(int i = 0; i < n_files; i++)
    {
      dir = entries[i];
      if(dir->d_type != DT_DIR) // regular file, fifo, etc
      {
        if(filename_matchExisting(dir->d_name, pattern))
          dirlist_addNew(plist, dir->d_name, dir->d_type);
      }
    }
  }
  else // no separation of dirs and files
  {
    for(int i = 0; i < n_files; i++)
    {
      dir = entries[i];
      if(filename_matchExisting(dir->d_name, pattern))
        dirlist_addNew(plist, dir->d_name, dir->d_type);
    }
  }

  // finish up, removing hidden and oddball files.
  if(sort_method & DL_NO_HIDDEN)
  {
    for(int i = dirlist_last(plist); i >= 0; i--)
      if(dirlist_getName(plist, i)[0] == '.')
        dirlist_remove(plist, i);
  }
  
  if(sort_method & DL_NO_IRREG)
  {
    int type;
    for(int i = dirlist_last(plist); i >= 0; i--)
    {
      type = dirlist_getType(plist, i);
      if(!(type & DT_REG) && !(type & DT_DIR))
        dirlist_remove(plist, i);
    }
  }  
  
  if(entries) free(entries);
  free(dirname);
  
  // return how many items remain.
  return dirlist_getCount(plist);
}


bool dirlist_isDir(DIRLIST* plist, int idx)
{
  if(dirlist_getType(plist, idx) == 4) // S_ISDIR
    return true;
  else
    return false;
} 

bool dirlist_isReg(DIRLIST* plist, int idx)
{
  if(dirlist_getType(plist, idx) == 8) // S_ISREG
    return true;
  else
    return false;
} 
 
