// filename_match.cpp - wildcard pattern checking for existing 
// files and for generating new names for renaming, or copying,
// also based on wildcard pattern criteria.

#include <stdio.h>		// fgets
#include <string.h>		// strlen()
#include <malloc.h>   // malloc

#include <LQ/filename.h>

#include <LQ/cdecls.h>
LQ_BEGIN_CDECLS;

// working buffers for matchNew
static char buf_in[FILENAME_MAXSTR];        // input name for matchNew
static char buf_out[FILENAME_MAXSTR];       // output name from matchNew

// returns true if string s matches pattern p which may include 
// wildcard chars '*' and '?'.  Handles escaped chars in pattern
// allowing avoidance of trigraph errors, etc.
static bool _match_existing (const char* s, const char* p)
{
  for (;;)
  {
    switch (*p++)
    {
      case '?':              // matches any character
          if (!*s++)
            return false;
          break;
          
      case '*':              // match 0-n of any characters
          if (!*p)           // end of wildcards?
            return true;     // matches all the rest
            while (!_match_existing(s, p))
              if (!*s++)
                return false;
              return true;
            
      case 0:                 // end of pattern
          return !*s;         // true if also end of s
          
      case '\\':              // next character literal
          if (*p)
            p++;              // falls through to default
      default:
        if (*s++ != *(p - 1)) // do letters match?
          return false;
        break;
    } // switch(*p++)
  }
}


/* based on part of copy300 by Marty Peritsky
License: GPL v.2+.
*/


static bool _match_new(const char *f, const char *p)
{
  int i;
  
  for(i = 0; i < FILENAME_MAXSTR; i++)
    buf_out[i] = '\0';

  i = 0;
  do{
    switch (*p){
      case '?':
      {
        switch (*f){
          case '.':
          case '\0':
            p++; 
            break;
          default:
            buf_out[i++] = *(f++); p++; 
            break;
        } // switch(*f)
        break;
      } // case '?'
        
      case '*':
      {
        switch (*f) {
          case '.':
          case '\0':
            p++; 
            break;
          default:
            do{ buf_out[i++] = *(f++);
            }while((*f != '.') && (*f != '\0') && (i < FILENAME_MAXSTR));
            do{ p++; 
            }while((*p != '.') && (*p != '\0'));
            break;
        } // switch(*f)
        break;
      } // case '*':
      case '.':
      {
        switch (*f) {
          case '.':
            f++;
          // falls through
          case '\0':
            buf_out[i++] = '.'; p++; 
            break;
          default:
            do{ f++; 
            }while((*f != '.') && (*f != '\0'));
            break;
        } // switch(*f)
        break;
      } // case '.'
      case '\0':
        break;
      default:
      {
        switch (*f)
        {
          case '.':
          case '\0':
            break;
          default:
            f++;
            break;
        } // switch(*f)
        buf_out[i++] = *(p++);
        break;
      } // default
    }
  }while(((*p) != '\0') && (i < FILENAME_MAXSTR));
  return *buf_out != 0;
}


bool filename_matchExisting(const char* name, const char* pattern)
{
  // save old name in case we also want a new name
  strcpy(buf_in, name);
  return _match_existing(name, pattern);
}

/* Returns true if any possibly quoted patterns in a comma separated list 
 match.  Each pattern is parsed according to the rules for filename_matchExisting().
 Speces before or after commas are ignored. -rs
*/

bool filename_matchAnyExisting(const char* name, const char* patterns)
{
  char tmp[256];
  strcpy(tmp, patterns);
  int len = strlen(tmp);
  
  // remove spaces before or after commas
  for(int i = 0; i < len; i++){
    if(tmp[i] == ' ' && tmp[i+1] == ','){
      memmove(tmp + i, tmp + i + 1, len - i);
      len --;
      i --;
    }
    else if(tmp[i] == ',' && tmp[i+1] == ' '){
      memmove(tmp + i + 1, tmp + i + 2, len - i -1);
      len--;
      i--;
    }
  }
  const char* seps = "\";,";
  char* p = strtok(tmp, seps);
  for(; p != 0; p = strtok(0, seps)){
    if(filename_matchExisting(name, p))
      return true;
  }
  return false;
}


// assumes filename_match as been run successfully first.  Sets name_out to name based on 
// replacement pattern and returns 0 if successful.
int filename_matchNew(char* name_out, size_t outlen, const char* name_in, const char* pattern)
{
  *name_out = 0;
  _match_new(name_in, pattern);
  if((*buf_out) && (strlen(buf_out) < outlen))
  {
    strcpy(name_out, buf_out);
    return 1;
  }
  return 0;
}

LQ_END_CDECLS;
