// objlist.c
#include <malloc.h>
#include <stdlib.h> // exit()

#include <LQ/objlist.h>

typedef struct
{
  const char* id;
  objlist_dtor_t dtor;
  long n_items;
}OBJLIST_HDR;

OBJLIST_HDR* base;
OBJLIST* current;
int current_ctr;

typedef struct
{
  const char* id;
  objlist_dtor_t dtor;
  long n_items;
  char* s1;
  char* s2;
  char* s3;
  char* s4;
}OBJLIST_VIEW;

OBJLIST_VIEW** view = (OBJLIST_VIEW**)&base;

// returns index of first item (always 0 if any items in the array, else -1.
int objlist_first(OBJLIST* plist)
{
  if(objlist_getCount(plist))
    return 0;
  else 
    return -1;
}

// returns index of last item (count -1), returns -1 if none.
int objlist_last(OBJLIST* plist)
{
  return objlist_getCount(plist) - 1;
}

// Sets item destructor field for the list so objlist can handle
// freeing items and arrays by calling back the appropriate function.
void objlist_setDestructor(OBJLIST* plist, objlist_dtor_t func)
{  
  base = ((OBJLIST_HDR*)*plist) - 1;
  base->dtor = func;
}

// for detecting valid objlist derivatives
static const char* objlist_ID = "objlist";

// for displaying errors and warnings, 1 = warn, 2 = fatal.
static void objlist_perror(int error_level)
{
  const char* msg = "%s [%s]: invalid object pointer\n";
  if(error_level == 2)
  {
    fprintf(stderr, msg, "Fatal", objlist_ID);
    exit(1);
  }
  else if(error_level == 1)
    fprintf(stderr, msg, "Warning", objlist_ID);
}

// for detecting imbalances in enter/leave pairs
static void objlist_nestError()
{
  fprintf(stderr, "Error: enter/leave in objlist imbalanced\n");
  exit(1);
}

// quick checks if new object pointer and is valid, returns 0 if ok
static int objlist_enter(OBJLIST* any, int error_level)
{
  if(current_ctr == 0)
    current = any;
  else if(current != any)
    objlist_nestError();
  current_ctr++;
  do
  {
    if(!any) break;
    if(!*any) break;
    base = ((OBJLIST_HDR*)(*any))-1;
    if(base->id != objlist_ID) break;
    return /* valid = */ 1;
  }while(0);
  // what to do if not valid
  if(error_level) objlist_perror(error_level);
  // else
  return 0; // false;
}

// unnests from objlist_enter() call
static void objlist_leave(OBJLIST* any)
{
  if(current != any)
    objlist_nestError();
  
  current_ctr--;
  if(current_ctr == 0) 
    current = 0;
}

// returns a pointer to the first (null) object in a new list.
OBJLIST objlist_new()
{
  base = (OBJLIST_HDR*)malloc(sizeof(OBJLIST_ITEM*) + sizeof(OBJLIST_HDR));
  base->id = objlist_ID;
  base->dtor = free; // default destructor
  base->n_items = 0;
  base[1].id = 0;
  return (OBJLIST_ITEM**)&base[1].id;
}

// only frees the object, does not adjust list count
static void objlist_freeItem(OBJLIST* plist, int item)
{
  OBJLIST_ITEM** list = *plist;
  base = ((OBJLIST_HDR*)list)-1;
  if(list[item])
    base->dtor((list)[item]);
  (list)[item] = 0;
}


// removes item and deletes the base
void objlist_delete(OBJLIST* plist)
{
  int i; // loop index
  objlist_enter(plist, 2);
  OBJLIST_ITEM** list = *plist;
  base = ((OBJLIST_HDR*)list) - 1;
  for(i = 0; i < base->n_items; i++)
  {
    if(list[i])
    {
      base->dtor(list[i]);
      base->n_items--;
    }
  }
  free(base);
  *plist = 0;
  objlist_leave(plist);
}

// returns item at index
OBJLIST_ITEM* objlist_get(OBJLIST* plist, int index)
{
  if(index >= objlist_getCount(plist))
    return 0;
  if(index < 0)
    return 0;
  return (*plist)[index];
}

// USE WITH CAUTION: sets item at index, does not free previous
void _objlist_set(OBJLIST* plist, int index, OBJLIST_ITEM* item)
{
  if(index >= objlist_getCount(plist))
    return;
  if(index < 0)
    return;
  (*plist)[index] = item;
}

int objlist_count(OBJLIST_ITEM** list)
{
  return ((OBJLIST_HDR*)list)[-1].n_items;
}

int objlist_getCount(OBJLIST* plist)
{
  return ((OBJLIST_HDR*)*plist)[-1].n_items;
}

static void objlist_setCount(OBJLIST* plist, int n)
{
  ((OBJLIST_HDR*)*plist)[-1].n_items = n;
}


// removes items, does not resize or delete the base
void objlist_clear(OBJLIST* plist)
{
  int i; // loop index
  objlist_enter(plist, 2);
  OBJLIST_ITEM** list = *plist;
  base = ((OBJLIST_HDR*)list) -1;
  for(i = 0; i < base->n_items; i++)
  {
    if(list[i])
      base->dtor(list[i]);
    list[i] = 0;
  }
  base->n_items = 0;
  *plist = list;
  objlist_leave(plist);
}

void objlist_append(OBJLIST* plist, OBJLIST_ITEM* o)
{
  objlist_enter(plist, 2); // 2 = fatal
  OBJLIST_ITEM** list = *plist;
  base = ((OBJLIST_HDR*)list) - 1;
  // add two below for new item plus terminator below
  int next_size = sizeof(OBJLIST_HDR) + (sizeof(OBJLIST_ITEM*) * (base->n_items + 2));
  base = (OBJLIST_HDR*)realloc(base, next_size);
  list = (OBJLIST_ITEM**)&base[1].id;
  list[base->n_items] = o;
  base->n_items++;
  list[base->n_items] = 0; // null last object, always
  *plist = list;
  objlist_leave(plist);
}

// removes (and frees) item at idx if any. Does not resize the list
void objlist_remove(OBJLIST* plist, int idx)
{
  int i; // loop index
  objlist_enter(plist, 2);
  OBJLIST_ITEM** list = *plist;
  int n_items = objlist_getCount(plist);
  if((idx < 0) || (idx > n_items))
  {
    objlist_leave(plist);
    return;
  }
  
  objlist_freeItem(plist, idx);
  
  for(i = idx; i < n_items; i++)
    list[i] = list[i+1];
  
  objlist_setCount(plist, n_items - 1);
  objlist_leave(plist);
}

// Can't do this without knowing size of the objects
/*
void objlist_addList(OBJLIST* plist1, OBJLIST* plist2)
{
  int i; // loop index
  objlist_enter(plist1, 2);
  
  base = ((OBJLIST_HDR*)*plist1) - 1;
  int n_items = objlist_getCount(plist1);
  int n_added = objlist_getCount(plist2);
  int new_total = n_items + n_added; 
  
  // +1 for terminating null
  int next_size = sizeof(OBJLIST_HDR) + sizeof(OBJLIST_ITEM*) * (new_total + 1);
  base = (OBJLIST_HDR*)realloc(base, next_size);
  
  // this just addresses the list area since id is the first field in the 
  // hdr and ends up being a convenient type.
  (*plist1) = (OBJLIST_ITEM**)&base[1].id;
  {  
    OBJLIST list1 = *plist1;
    OBJLIST list2 = *plist2;
    for(i = 0; i < n_added; i++)
      list1[n_items + i] = list2[i];
  }
  objlist_setCount(plist1, new_total);
  // rs
  // *plist1[new_total] = 0;
  (*plist1)[new_total] = 0;
  objlist_leave(plist1);
}
*/

// inserts new item at idx, or end or at beginning if out of range
void objlist_insert(OBJLIST* plist, OBJLIST_ITEM* o, int idx)
{ 
  int i; // loop index 
  // easiest to append and then to move
  objlist_append(plist, o);
  
  OBJLIST_ITEM** list = *plist;  
  int n_items = objlist_count(list);
  
  // check range 
  if(idx < 0)
    idx = 0;
  else if(idx >= n_items)
    idx = n_items - 1;
  
  if(idx == n_items - 1) 
    return;
  
  // grab the appended object, then shift the list up
  // and put the new object in at idx.
  void* tmp = list[n_items - 1];
  
  // point to last and copy prev to current and decrement
  for(i = n_items - 1; i > idx; i--)
    list[i] = list[i-1];
  list[idx] = tmp;
}

// removes to end from idx, inclusive
void objlist_trimRight(OBJLIST* plist, int idx)
{
  int i; // loop index
  OBJLIST_ITEM** list = *plist;
  int n_items = objlist_count(list);
  if(idx >= n_items)
    return;
  if(idx < 0)
    idx = 0;
  for(i = idx; i < n_items; i++)
    objlist_freeItem(&list, i);
  
  objlist_setCount(&list, idx);
  *plist = list;
}

// removes from beginning to idx, inclusive
void objlist_trimLeft(OBJLIST* plist, int idx)
{
  if(idx < 0)
    return;
  OBJLIST_ITEM** list = *plist;
  int n_items = objlist_count(list);
  
  // don't resize but step through the objects and delete
  // and move from src to dest as we go
  int dest = 0;
  int src = 0;
  while(src <= idx)
    objlist_freeItem(&list, src++);
  
  while(src < n_items)
  {
    list[dest++] = list[src];
    list[src++] = 0;
  }
  objlist_setCount(&list, dest);
  *plist = list;
}
