// makefile creator mc2 generates working makefiles using mc2.def makefile defs

#include <stdio.h>          // printf, sprintf, etc.
#include <stdlib.h>         // exit(), system()
#include <unistd.h>         // getpid() and tons of stuff
#include <string.h>         // strrchr(), etc.
#include <errno.h>          // errno
#include <ctype.h>          // toupper()

#include <LQ/slist.h>       // string list functions
#include <LQ/filename.h>    // filename_exists()
#include <LQ/dirlist.h>     // dirlist funcs
#include <LQ/unparse_args.h>
#include <LQ/getch-term.h>  // getch() and other keyboard stuff
#include <LQ/tmpstr.h>      // tmpstr()
#include "utils.h"          // have_mc2_def()
#include "mc2-parser.h"     // parse_mc2_file()
#include "mc2-writer.h"     // MC2_VARS, 

void dbg() {} // a breakpoint for kdbg.exec

#define streq(a, b) (strcmp(a, b) == 0)

// vars controling what gets written
int mc2_def_only = 0; // if true writes mc2.def instead of makefile

// returns error code
int usage(int errcode);
int short_usage(int errcode);
int version(int errcode);

/// handles commandline switches, exits or returns,
/// may set env vars or call -run <prog> or -create
/// <fileset> as well as the normal switches.
void do_switches(int argc, char** argv);

extern void tst();

int main(int argc, char **argv)
{
  dbg();
  tst();
  
  // an object containing all the vars needed by the writer
  MC2_VARS vars;
  
  int err = 0;
  
  // init dirs and and app name
  appname = basename(argv[0]);
  setExecDir(mc2_dir, mc2_dir_max);
  setenv("MC2_DIR", mc2_dir, 1);
  pwd = getenv("PWD");  
  
  // commandline switches, if it returns, we continue.
  do_switches(argc, argv);
  
  // lists for names of sources, headers and object files for
  // 'main' type Makefiles, and subdirs for 'project' type.
  char **srclist = slist_new();
  char **hdrlist = slist_new();
  char **objlist = slist_new();
  char **subdirs = slist_new();
  
  // load the vars object
  vars.psrclist = &srclist;
  vars.phdrlist = &hdrlist;
  vars.pobjlist = &objlist;
  vars.psubdirs = &subdirs;
  
  err = create_subdirlist(&subdirs);    

  // load env vars from mc2.def file if undefined
  dbg();
  if(have_mc2_def())    
    parse_mc2_file(&subdirs); 

  // finish the vars struct variable definitions, loading
  // overrides if defined at this point.
  
  // compiler name for 'main' type makefiles if none set yet
  const char *default_compiler_name = "gcc";
  
  vars.outname = check_override("OUTNAME", "main");
  vars.srcdir = check_override("SRCDIR", "src");
  vars.objdir = check_override("OBJDIR", "o");
  vars.bindir = check_override("BINDIR", ".");
  vars.compiler_name = check_override("COMPILER_NAME", default_compiler_name);
    
  // create any missing folders and load file lists if 'main' type
  if(!streq(vars.outname, "PROJECT") && !streq(vars.outname, "NONE"))
    setup_main(&vars);
    
  ////////////////////////////////
  // all set, now this is where the action is...
  if(streq(vars.outname, "NONE"))
  {
    int ack = 0;
    if(!filename_exists("Makefile"))
    {
      printf("No pre-existing Makefile yet.  Skipping...\n");
      ack++;
    } // doesn't exist
    else
    {
      // the Notice has already been printed out... add this.
       printf("For using pre-existing makefiles in mc2 PROJECT builds\n");
     // if the makefile doesn't include a tag named 'update:' issue
     // a warning and wait for a response.
     write_NONE_type();
     if(!find_target("update:", "Makefile"))
     {
       fprint_hline(stdout, '=', 60);
       printf("%s", 
        "There is no target named 'update:' in the current Makefile\n"
        "This will cause problems for 'make update' in the PROJECT\n"
        "Makefile.  You can manually add\n"
        "   'update:'\n"
        "followed by an empty line at the bottom of in the old Make-\n"
        "file to enable this to become an mc2 PROJECT inclusion.\n"
        );
       fprint_hline(stdout, '=', 60);
       ack++;
     } // missing 'update target
    } // else
    if(ack)
    {
      printf("  Press ENTER to acknowledge.\n");
      getchar();
    } // ack
  } // type NONE or PROJECT
  else if(mc2_def_only)
    err = write_mc2_def(&vars);
  else
  {
    err = write_makefile(&vars);
  }
    
  if(err)
    doerr("Can't write makefile");
    
  // clean up
  slist_delete(&srclist);
  slist_delete(&hdrlist);
  slist_delete(&objlist);
  slist_delete(&subdirs);

  return 0;
}


// using an slist object to communicate with an xmessage dialog
void usageDlg(const char* helpstr)
{
  char* msg = tmpstr(strlen(helpstr) + 256, "");
  sprintf(msg, helpstr, appname);
  char* res;
  // returns a tmpstr()
  res = msgbox("OK,STDOUT", msg, "-default OK");
  
  if(streq(res, "STDOUT"))
    printf(strchr(msg, ']') + 2 );
}

int usage(int errcode)
{
#include "usage_str.dat"
  usageDlg(usage_str);
  return errcode;
}

int short_usage(int errcode)
{
#include "short_usage_str.dat"
  usageDlg(short_usage_str);
  return errcode;
}

int version(int errcode)
{
#include "version.h"
  printf("%s version %s", appname, VERSION);
  return errcode;
}

//////////////////////////////////////////////////
// switch-funcs.cpp

typedef int (*tokfn_t)(int argc, char** argv);
typedef struct 
{
  const char* name;
  tokfn_t fn;
}toktable;

// external functions

// always returns 1
int ext_fileError(const char* path)
{
  char* dir = strdup(path);
  char* file = strrchr(dir, '/');
  if(file) 
    *file++ = 0;
  else
    file = (char*)"???";
  
  fprintf(stderr, "The file '%s' does not exist in directory\n%s\n", file, dir);
  free(dir);
  return 1;
}

int ext_run(int argc, char** argv)
{
  if(argc < 2)
    return ext_fileError("MC2_DIR/run/[NO NAME]");
  char * cmd = tmpstr(256, 0);
  char * params = tmpstr(256, 0);
  // test the file part of the command first, does the file exist to run it?
  sprintf(cmd, "%s/run/%s", mc2_dir, argv[1]);
  if(!filename_exists(cmd))
    return ext_fileError(cmd);
  unparse_args(params, 256, argc - 1, argv + 1);  
  sprintf(cmd, "%s/run/%s %s", mc2_dir, argv[1], params);
  system(cmd);
  return 0;
}

// returns 0 on success
int ext_create(int argc, char** argv)
{
  dbg();
  if(argc < 2)
    return short_usage(1);
  
  char* cmd = tmpstr(256, 0);
  char* params = tmpstr(256, 0);
  // test the file part of the command first, does the file exist to run it?
  unparse_args(params, 256, argc - 1, argv + 1);
  sprintf(cmd, "%s/create/%s/init.sh", mc2_dir, argv[1]);
  
  if(filename_exists(cmd))
  {
    sprintf(cmd + strlen(cmd), " %s", params);
    return system(cmd);
  }
  
  // else
  sprintf(cmd, "cp -rf %s/create/%s/* .", mc2_dir, argv[1]);
  if(system(cmd))
  {
    printf("Error copying files from '%s'\n", cmd + 7);
    return 1;
  }
  return 0;
}
    

toktable ext_commands[] =
{
  {"-run",      ext_run},
  {"-create",   (tokfn_t)ext_create},
  {0, 0}
};


// internal commands

int int_help(int nparams, char** params) 
{ 
  if(nparams == 1) 
    exit(short_usage(0));
  else
    exit(usage(0));
}

int int_version(int nparams, char** params) {  exit(version(0)); }

// returns -1 (continue)
int int_t(int nparams, char** params)
{
  int offset = 1;
  if(strcasecmp(params[offset], "c") == 0)
    setenv("COMPILER_NAME", "gcc", 1);
  else if(strcasecmp(params[offset], "c++") == 0)
    setenv("COMPILER_NAME", "g++", 1);
  else 
  {
    setenv("COMPILER_NAME", "gcc", 1);
    offset--;
  }
  offset ++;
  
  if(offset < nparams)
  {
    setenv("OUTNAME", params[offset], 1);
    // if capitalized
    if(!strcmp(params[offset],"PROJECT") || !strcmp(params[offset], "MULTI") || !strcmp(params[offset], "NONE"))
    {
      printf("Notice: creating '%s' type mc2.defs\n", params[offset]);
      if(!filename_exists("Makefile"))
        system("mc2 -init");
    }
    // oddball case
    else if(!strcasecmp(params[offset],"project") || !strcasecmp(params[offset], "multi") || !strcasecmp(params[offset], "none"))
    {
      printf("Notice: creating mc2.defs with OUTNAME = '%s'\n", params[offset]);
      printf("  (If you intended to create a PROJECT, MULTI, or NONE type\n"
             "  makefile, run this again with the <outname> parameter in CAPS\n"
             "  Continuing operation as previously specified.)\n");
    }
    offset++;
  }
  mc2_def_only = 1;
  return -1; // continue
}


// list templates by extension
int int_list(int argc, char** argv)
{
  char* p = argv[1];
  const char* default_dir = "templates";
  const char* run_dir = "run";
  const char* create_dir = "create";
  const char* filterstr = "mc2.*";
  const char* listing_dir;
  int file_type = 8; // S_IFREG
  
  if(argc == 1)
    listing_dir = default_dir;
  else
  {
    if(argv[1][0] == 'r')
    {
      listing_dir = run_dir;
      filterstr = "*";
    }
    else if(argv[1][0] == 'c')
    {
      listing_dir = create_dir;
      file_type = 4; // S_IFDIR
      filterstr = "*";
    }
    else
      listing_dir = default_dir;
  }
        
  char tmp[256];
  sprintf(tmp, "%s/%s", mc2_dir, listing_dir);
  DIRLIST_ITEM** list = dirlist_new();
  int nfiles = dirlist_read(&list, tmp, filterstr, -1);
  fprint_hline(stdout, '=', 64);
  
  // filter out dirs or files, depending...
  for(int i = nfiles - 1; i >= 0; --i)
  {
    if(list[i]->type != file_type)
      dirlist_remove(&list, i);
  }
  
  nfiles = dirlist_getCount(&list);
  
  printf("\n");
  
  if(listing_dir == default_dir)
  {
    printf("Saved templates (use mc2 -fetch <name> to load):\n");
  }
  else if (listing_dir == run_dir)
    printf("Programs for -run switch:\n");
  else if(listing_dir == create_dir)
    printf("File sets for -create switch:\n");  
  
  fprint_hline(stdout, '-', 64);
  for(int i = 0; i < nfiles; i++)
  {
    strcpy(tmp, dirlist_get(&list, i)->name);    
    if(listing_dir == default_dir)
    {
      // just print out the extension
      p = strchr(tmp, '.');
      if(p)   p++;
      else    p = tmp;
    }
    else
      p = tmp;
    
    printf("%-32s", p);
    
    if((i % 2) == 1)
      printf("\n");
  }
  printf("\n");
  fprint_hline(stdout, '=', 64);
  dirlist_delete(&list);
  return 0;
}


int int_license(int argc, char** argv)
{
  char* cmd = tmpstr(256, 0);
  sprintf(cmd, "xmessage -buttons \"OK\" -default \"OK\" -center -file '%s/GPL-LICENSE'", mc2_dir);
  system(cmd);
  return 0;
}

static int save_fetch_error(int errcode, const char* cmd)
{
  printf(
      "'mc2 %s' requires the name of a template\n"
      "Try 'mc2 -list' to view existing templates\n", cmd);
  return 1;
}

int choose_YN()
{
  term_setNonCanonical();
  term_setEcho(0);
  term_setCBreak(0); // disable ctrl-c while terminal is non-canonical
  int ch = 0;
  for(;;)
  {
    ch = toupper(getch());
    if((ch == 'N') || (ch == 'Y'))
    {
      printf("%c\n", ch);
      break;
    }
    else 
      printf("\nPlease select either Y or N\n");
  }
  term_setCanonical();
  return ch;
}

// save current mc2.def file as a new template
int int_save(int argc, char** argv)
{
  int ch = 'Y';
  char* syscmd;
  
  if(argc != 2)
    return save_fetch_error(1, argv[0]);
  
  if(filename_exists(tmpstr(256, "%s/templates/mc2.%s", mc2_dir, argv[1])))
  {
    printf("Overwrite existing file? [Y/N]: ");
    fflush(stdout);
    ch = choose_YN();
  }
     
  if(ch == 'Y')
  {
    printf("Writing new template '%s'\n", argv[1]);
    syscmd = tmpstr(256, "cp mc2.def %s/templates/mc2.%s", mc2_dir, argv[1]);
    system(syscmd);
  }
  else
    printf("Existing template unchanged\n");
  return 0;
}

// fetch an existing mc2.def template for current build
int int_fetch(int argc, char** argv)
{
  int ch = 'Y';
  char* syscmd;
  
  if(argc != 2)
    return save_fetch_error(1, argv[0]);
  
  if(filename_exists(tmpstr(256, "mc2.def", mc2_dir, argv[1])))
  {
      printf("Overwrite existing mc2.def file? [Y/N]: ");
      fflush(stdout);
      ch = choose_YN();
  }
     
  if(ch == 'Y')
  {
    syscmd = tmpstr(256, "cp %s/templates/mc2.%s mc2.def", mc2_dir, argv[1]);    
    // if exists
    if(!system(syscmd))
    {
      printf("Writing new mc2.def in current directory\n");
      if(!filename_exists("Makefile"))
      {
        system("mc2 -init");
        printf("New Makefile initialized.\n");
      }
      else
        printf("Existing Makefile unchanged.\n");
    }
    else // doesn't exist
      printf("Can't write mc2.def\n");
  }
  else
    printf("Existing file unchanged\n");
  return 0;
}

// update an existing mc2.def template for current build
int int_update(int argc, char** argv)
{ 
  if(have_mc2_def()); 
  printf("Updating/Creating Makefile from mc2.def file\n");
  return -1; // just continue
}

// remove an existing mc2.def template
int int_remove(int argc, char** argv)
{
  char* filename;
  if(argc == 1)
  {
    printf("No template name, can't remove template.\n");
    return(1);
  }
  if(argc > 2)
  {
    printf("Can't remove more than one template at a time.\n Aborting.\n");
    return(1);
  }
  filename = tmpstr(256, "%s/templates/mc2.%s", mc2_dir, argv[1]);
  if(!filename_exists(filename))
  {
    printf("Can't find template (extenstion) '%s'\n", argv[1]);
    return(0);
  }
  remove(filename);
  return(0);
}

toktable int_commands[] =
{
  {"-help",      (tokfn_t)int_help},
  {"-v",         (tokfn_t)int_version},
  {"-version",   (tokfn_t)int_version},
  {"-t",         (tokfn_t)int_t},
  {"-list",      (tokfn_t)int_list},
  {"-license",   (tokfn_t)int_license},
  {"-save",      (tokfn_t)int_save},
  {"-fetch",     (tokfn_t)int_fetch},
  {"-remove",    (tokfn_t)int_remove},
  {"-update",    (tokfn_t)int_update},
  {"-u",         (tokfn_t)int_update},
  {"-init",      (tokfn_t)int_update},
  {"-i",         (tokfn_t)int_update},
{0, 0}
};


int find_tok(toktable* toks, char* name)
{
  // make '--' swtiches '-' switches
  const char *find = name;
  if(find[1] == '-')
    find++;
        
  for(int i = 0; toks[i].name != 0; i++)
  {
    if(streq(toks[i].name, find))
      return i;
  }
  return -1;
}


/// handles commandline switches, exits or returns
void do_switches(int argc, char** argv)
{
  // or show usage (and exit), or do a setup (and continue).
  if(argc == 1)
  {
    if(have_mc2_def())
    {
      printf("An 'mc2.def' file exists, did you mean to -u[pdate] the Makefile?\n Aborting.\n");
      exit(0);
    }
    else
      exit(short_usage(1));
  }

  int tok;
  int retcode;
  
  for(int i = 1; i < argc; i++)
  { 
    if(argv[i][0] == '-')
    {    
      tok = find_tok(ext_commands, argv[i]);
      if(tok >= 0)
      {
        tokfn_t tokfn = ext_commands[tok].fn;
        retcode = tokfn(argc - i, argv + i);
        if(retcode >= 0)
          exit(retcode);
        else 
          return; 
      }
      tok = find_tok(int_commands, argv[i]);
      if(tok >= 0)
      {
        tokfn_t tokfn = int_commands[tok].fn;
        retcode = tokfn(argc - i, argv + i);
        if(retcode >= 0)
          exit(retcode);
        else 
          return; 
      }
      // else
      fprintf(stderr, "Unknown switch '%s'\n", argv[i]);
      exit(1);
    }    
  }  
}

