#include <ctype.h>    // isalpha(), ...
#include <string.h>   // strcmp()
#include <stdio.h>    // printf()
#include <stdlib.h>   // setenv()

#include <LQ/slist.h>
#include <LQ/filename.h>

#define NO_PARSER_WARN 1
#include <LQ/Parser.h> // include last

#define streq(a, b) (strcmp(a, b) == 0)

using namespace Parser;  // begin Parser 

// custom parser funcs

bool sws() // skip optional whitespaces
{
  while(obj_skip_ws())
    ; // noop
  return obj.state = true;
}

bool skip_empty_line()
{
  SAVE_PTRS();
  sws();
  if(!(obj_skip("\n") || obj_is_eoi()))
    RESTORE_PTRS();
  return obj.state;
}

bool is_empty_line()
{
  SAVE_PTRS();
  sws();
  if(obj_is_eoi() || obj_is("\n"))
    ; // noop
  RESTORE_PTRS();
  return obj.state;
}

bool skip_comment()
{
  SAVE_PTRS();
  sws();
  if(obj_is("#"))
  {
    obj_skip_linetext();
    obj_skip("\n");
    return obj.state = true;
  }
  else return obj.state = false;
}

bool copy_name()
{
  // alpha or underscore followed by alphanum or underscore
  // to end
  if((!isalpha(*obj.ip) || obj_is("_")))
    return obj.state = false;
  obj_copy_nchars(1);
  while(isalnum(*obj.ip) || obj_is("_"))
    obj_copy_nchars(1);
  return obj.state = true;
}

static char assignment_name[256];
static int override_index;

static const char* override_names[] =
{
  "OUTNAME",
  "SRCDIR",
  "OBJDIR",
  "BINDIR",
  "COMPILE",
  "CFLAGS",
  "INCLUDE",
  "LINK",
  "LDFLAGS",
  "LIB",
  0
};

// returns index of overrid or -1 if not found
static int find_override(const char* name)
{
  for(int i = 0; override_names[i] != 0; i++)
  {
    if(streq(override_names[i], name))
      return i;
  }
  return -1;
}


bool is_assignment()
{
  char* opname;
  SAVE_PTRS();  
  sws();
  opname = obj.op;
  if(copy_name())
  {
    strcpy(assignment_name, opname);
    sws();
    obj_is("=");
  }
  
  RESTORE_PTRS();
  return obj.state;
}

bool do_assignment(char*** list)
{
  char* opmark = obj.op;
  *opmark = 0;
  
  obj_copy_to("\n"); // allows backslash escaped newlines
  obj_copy("\n");
  slist_append(list, opmark);
  obj_delete_op(opmark, obj.op);
  return obj.state = true; // redundant but harmless
}

// is_assignment() must be done first
bool is_override()
{
  is_assignment();
  int indx = find_override(assignment_name);
  if(indx >= 0)
    obj.state = true;
  else
    obj.state = false;
  override_index = indx;
  return obj.state;
}

// used in value strings to remove spaces, especially before 
// comments, that can cause 'make' to misread paths.
void trim_value(char* valuestr)
{
  // find first space on line before comment
  // and terminate the string there.
  char* p = valuestr + strlen(valuestr) - 1;
  while(strchr(" \t", *p) && p >= valuestr)
    p--;
  
  // queue up the first blank, if any found
  p++;
  
  // terminate the value part of the string
  *p = 0;
}

// used in value strings
void remove_comments(char* valuestr)
{
  // if a value has a comment in it, remove it.
  
  char* p = strchr(valuestr, '#');
  if(p) *p = 0;
}                     


bool do_override()
{
  char* opmark = obj.op;
  char* newvalue;
  
  obj_skip_to("=");
  if(!obj.state)
    printf("this is impossible!! :-) \n");
  obj_skip_nchars(1);
  sws();
  newvalue = obj.op;
  obj_copy_linetext(); 
  remove_comments(newvalue);
  trim_value(newvalue);
  setenv(override_names[override_index], newvalue, 0);
  obj.op = opmark; // delete
  return obj.state = true;
}

bool is_mkinclude()
{
  SAVE_PTRS();  
  sws();
  obj_is("include");
  RESTORE_PTRS();
  return obj.state;  
}

// also goes into mk_head list verbatim so the name is just to
// (hopefully) avoid confusion.  It's the same as do_assignment
bool do_mkinclude(char*** list)
{
  do_assignment(list);
  slist_append(list, "\n"); // just an additional newline separator
  return obj.state = true;
}

bool is_target()
{
  // contains a colon anywhere on the line
  SAVE_PTRS();
  char* opmark = obj.op;
  obj_copy_linetext();
  if(strchr(opmark, ':'))
    obj.state = true;
  else
    obj.state = false;
  
  // undo
  RESTORE_PTRS();
  return obj.state;  
}


bool is_subdirs()
{
  SAVE_PTRS();
  sws();
  obj_skip("@");
  obj_is("SUBDIRS ");
  RESTORE_PTRS();
  return obj.state;
}

bool do_subdirs_command(char*** psubdirlist)
{
  char cmd[256];
  char buf[256];
  char** subdirlist = *psubdirlist;
  if(!slist_count(subdirlist))
  {
    obj_skip_linetext();
    obj_skip("\n");
    return obj.state = true; // not an error
  }
  char* opmark = obj.op;
  bool has_at;
  obj_copy_to("SUBDIRS");
  has_at = (strchr(opmark, '@') != 0 ? true : false);
  obj_skip_to(" "); // checked for ws in is_subdirs()
  sws();
  
  // the rest of the line is the command to repeat for every subdir
  obj.op = opmark;
  obj_copy_linetext();
  obj_skip("\n");
  strcpy(cmd, opmark);
  obj.op = opmark;
  for(int i = 0; i < slist_count(subdirlist); i++)
  {
    sprintf(buf, "\t%scd %s && %s\n",
    has_at ? "@" : "", (subdirlist)[i], cmd);
    obj_put(buf);
  }
  return obj.state = true;
}

bool do_target(char*** list, char*** psubdirlist)
{
  // TODO check if target is a dup and if it is, insert 'old_' 
  // in front of previous version.
  
  char* opmark = obj.op;
  *opmark = 0;
  // copy first line as is
  obj_copy_to("\n");
  obj_copy("\n");
  
  for(;;)
  {
    // then, until empty line, insert TAB, skip whitespace and copy rest of line.
    // and if first word is SUBDIR, cd and copy the command for every subdirectory.
    if(obj_is_eoi() || is_empty_line())
    {
      obj_put("\n"); // just a newline
      break;
    }
    
    if(is_subdirs())
      do_subdirs_command(psubdirlist);
    else
    {
      obj_put("\t");
      sws();
      obj_copy_linetext();
      obj_copy("\n");
    }
  }
  slist_append(list, opmark);
  
  // delete
  obj_delete_op(opmark, obj.op);
  return obj.state = true;
}

void make_op_str(char** list)
{
  int nlines = slist_count(list);
  obj_put("\n");
  for(int i = 0; i < nlines; i++)
  {
    obj_put(list[i]);
  }
}

void parse_mc2_file(char*** psubdirlist)
{
  size_t buflen;
  char* buf;
  bool state = filename_load("mc2.def", &buf, &buflen);
  if(!state) return; // shouldn't happen
  
  char** mk_head = slist_new();
  char** mk_foot = slist_new();
  
  // read a line
  // if comment skip line
  // if 'include' copy literally
  // if assignment = 
  //    if override, add to env if not already set, else add whole line+ to mk_head
  // if assignment :
  // set has_targets flag (to clear default) and add whole line to mk_foot and
  //    continue with TAB-ed lines until empty line.
  
  PARSER_OBJ This;
  parser_init(&This, buf, buflen, false);
  
  while(obj_not_eoi())
  {
    if(!((skip_empty_line() || skip_comment())))
    {
      if(is_assignment())
      {
        if(is_override())
          do_override();
        else
          do_assignment(&mk_head);
      }
      else if(is_mkinclude())
        do_mkinclude(&mk_head);
      else if(is_target())
        do_target(&mk_foot, psubdirlist);
      else
        break;
    }
  }
  if(obj_not_eoi())
  {
    char* opmark = obj.op;
    int nlines = 0;
    
    for(const char* p = obj.ibuf; p <= obj.ip; p++)
      if(*p == '\n')
        nlines++;
      
    printf("Syntax error: mc2.def line %d", nlines);
    obj_copy_linetext();
    printf("> %s", opmark);
    exit(1);
  }
  {
    obj.op = obj.obuf; // reset
    make_op_str(mk_head);
    setenv("MK_HEAD", obj_get_result(), 0);
    
    obj.op = obj.obuf; // reset
    make_op_str(mk_foot);
    setenv("MK_FOOT", obj_get_result(), 0);
  }
  free(buf);
  parser_free(&This);
  slist_delete(&mk_head);
  slist_delete(&mk_foot);
}

using namespace std; // end Parser


