
#include <stdlib.h> // getenv()
#include <string.h> // strcpy()

#include <LQ/slist.h>

#include "mc2-writer.h" 
#include "version.h" // VERSION
#include "utils.h"   // create_dirs

#define streq(a, b) (strcmp(a, b) == 0)


void write_makehdr(FILE * fp)
{
  fprintf(fp, 
          "## Makefile created with makefile creator 'mc2' v %s\n", VERSION
         );
}

void write_main_mod_section(FILE* fp)
{
  fprintf(fp, "\n");
  fprint_hline(fp, '#', 64);
  fprintf(fp,
              "## Note: The following section can be put into a 'mc2.def' and \n"
              "## modified for automatic override of defaults as needed.\n"
         );
  fprint_hline(fp, '#', 64);
  fprintf(fp, "\n");
}

void write_divider(FILE* fp)
{
  fprintf(fp, " \n");
  fprint_hline(fp, '#', 64);
  fprintf(fp, " \n");
}

void write_mk_head(FILE * fp)
{
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## Variable Definitions Section\n\n");
  
  if(getenv("MK_HEAD"))
  {
    fprintf(fp, "## User Defined Vars\n");
    fprintf(fp, getenv("MK_HEAD"));
  }
}

void write_assign_main_outname(FILE * fp, MC2_VARS* o)
{
  fprintf(fp,
          "## The output file name not including path\n"
              "OUTNAME = %s\n\n", o->outname);
}

void write_assign_main_dirs(FILE * fp, MC2_VARS* o)
{
  fprintf(fp,
          "## The directories for sources, (temp) objects, and binary output(s)\n"
              "BINDIR = %s\n"
              "SRCDIR = %s\n" "OBJDIR = %s\n\n", o->bindir, o->srcdir, o->objdir);
}


void write_assign_main_compiler(FILE * fp, MC2_VARS* o)
{
  // init defaults
  char compile[512];
  char cflags[512];
  char include[512];
  sprintf(compile, "%s -m32 -c -o", o->compiler_name);
  sprintf(cflags, "-O2 # or -g3");
  sprintf(include, "-I/usr/include -I./$(SRCDIR)");
  
  // use overrides if any
  strcpy(compile, check_override("COMPILE", compile));
  strcpy(cflags, check_override("CFLAGS", cflags));
  strcpy(include, check_override("INCLUDE", include));
  
  // don't add switch for asm, check for 'as ' as first chars
  if(compile != strstr(compile, "as "))
    check_end_switch(compile, "-c", "-o");
  
  fprintf(fp,
          "## What COMPILE should do.\n"
              "COMPILE = %s\n"
              "CFLAGS = %s\n" "INCLUDE = %s\n\n", compile, cflags, include);
}

// compiler name = linker name here.
void write_assign_main_linker(FILE * fp, MC2_VARS* o)
{
  const char* linker_name = o->compiler_name;
  // init defaults
  char link[512];
  char ldflags[512];
  char lib[512];
  sprintf(link, "%s -m32 -o", linker_name);
  sprintf(ldflags, "# -lc");
  sprintf(lib, "-L/usr/lib");
  
  // use overrides if any
  strcpy(link, check_override("LINK", link));
  strcpy(ldflags, check_override("LDFLAGS", ldflags));
  strcpy(lib, check_override("LIB", lib));
  
  // don't add switch for static lib, check for 'ar ' as first chars
  if(link != strstr(link, "ar "))
    check_end_switch(link, NULL, "-o");
  
  fprintf(fp,
          "## What LINK should do.\n"
              "LINK = %s\n" "LDFLAGS = %s\n" "LIB = %s\n\n", link, ldflags, lib);
}

void write_list(FILE * fp, const char *dir, char ***plist)
{
  int i;
  for(i = 0; i < slist_getCount(plist); i++)
  {
    fprintf(fp, "  %s/%s \\\n", dir, (*plist)[i]);
  }
}

void write_main_list_section(FILE * fp, MC2_VARS* o)
{
  fprintf(fp, "## MODIFY BELOW THIS LINE WITH GREAT CARE\n");
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## File Lists Section\n\n");
  fprintf(fp,
          "## The full path to the output file\n"
              "MAIN = $(BINDIR)/$(OUTNAME)\n"
              "\n");
           
  const char *endlist = "  #############\n\n";
           
  fprintf(fp, "SRC = \\\n");
  write_list(fp, o->srcdir, o->psrclist);
  fprintf(fp, endlist);
           
  fprintf(fp, "# If your your headers aren't all in SRCDIR, you can define\n");
  fprintf(fp, "# EXT_HDR in a makeinclude to add others to the dependencies\n");
  fprintf(fp, "HDR = \\\n");
  write_list(fp, o->srcdir, o->phdrlist);
  fprintf(fp, "  $(EXT_HDR) \\\n");
  fprintf(fp, endlist);
           
  fprintf(fp, "OBJ = \\\n");
  write_list(fp, o->objdir, o->pobjlist);
  fprintf(fp, "  $(EXT_OBJ) \\\n");
  fprintf(fp, endlist);
}


void write_project_file_lists(FILE* fp, MC2_VARS* o)
{
  const char *endlist = "  #############\n\n";
  
  fprintf(fp, "## MODIFY BELOW THIS LINE WITH GREAT CARE\n");
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## File Lists Section\n\n");
  
  fprintf(fp, "MC2_FILES = \\\n");
  for(int i = 0; i < slist_getCount(o->psubdirs); i++)
    fprintf(fp, "  %s/mc2.def\\\n", (*o->psubdirs)[i]);
  fprintf(fp, endlist);
  
  fprintf(fp, "PROJECT_FILES = \\\n");
  for(int i = 0; i < slist_getCount(o->psubdirs); i++)
    fprintf(fp, "  %s/Makefile\\\n", (*o->psubdirs)[i]);
  fprintf(fp, endlist);

}

static void _objlist2mainlist(char*** pmainlist, char***pobjlist)
{
  char tmp[256];
  char* p;
  // copy mains from objlist without .o
  for(int i = 0; i < slist_getCount(pobjlist); i++)
  {
    strcpy(tmp, (*pobjlist)[i]);
    p = strrchr(tmp, '.');
    if(!p)
    {
      fprintf(stderr, "Fatal error in objlist/mainlist translation\n");
      exit(1);
    }
    *p = 0;
    slist_append(pmainlist, tmp);
  }
}

void write_multi_list_section(FILE * fp, MC2_VARS* o)
{
  char** mainlist = slist_new();
  _objlist2mainlist(&mainlist, o->pobjlist);
  
  char** srclist = *o->psrclist;
  char** objlist = *o->pobjlist;
  char** hdrlist = *o->phdrlist;
    
  fprintf(fp, "## MODIFY BELOW THIS LINE WITH GREAT CARE\n");
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## File Lists Section\n\n");
  fprintf(fp,
          "## The full path to the output file\n"
              "MAIN = $(MAIN_FILES)\n"
              "\n");
           
  const char *endlist = "  #############\n\n";
           
  fprintf(fp, "SRC = \\\n");
  write_list(fp, o->srcdir, &srclist);
  fprintf(fp, endlist);
           
  fprintf(fp, "HDR = \\\n");
  write_list(fp, o->srcdir, &hdrlist);
  fprintf(fp, endlist);
           
  fprintf(fp, "OBJ = \\\n");
  write_list(fp, o->objdir, &objlist);
  fprintf(fp, endlist);
  
  fprintf(fp, "MAIN_FILES = \\\n");
  write_list(fp, o->bindir, &mainlist);
  fprintf(fp, endlist);
  
  slist_delete(&mainlist);
}

void write_main_rules(FILE * fp, MC2_VARS* o)
{
  char** srclist = *o->psrclist;
  char** objlist = *o->pobjlist;
  
  int i;      // loop counter
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## Rules Section\n");
  fprintf(fp, "\n");
  fprintf(fp, "all: $(MAIN) $(HDR) $(OBJ) $(SRC)\n");
  fprintf(fp, "\n");
  fprintf(fp, "$(MAIN): $(OBJ) $(HDR) $(SRC)\n");
  fprintf(fp, "\t@echo\n");
  fprintf(fp, "\t@echo \"Linking $(OUTNAME)\"\n");
  fprintf(fp, "\t$(LINK) $(MAIN) $(OBJ) $(LDFLAGS) $(LIB)\n");
  fprintf(fp, "\t$(POST)\n");
  fprintf(fp, "\n");
  for(i = 0; i < slist_count(srclist); i++)
  {
    fprintf(fp, "$(OBJDIR)/%s: $(SRCDIR)/%s $(HDR)\n", objlist[i],
            srclist[i]);
    fprintf(fp, "\t@echo\n");
    fprintf(fp, "\t@echo \"Compiling %s\"\n", objlist[i]);
    fprintf(fp,
            "\t$(COMPILE) $(OBJDIR)/%s $(SRCDIR)/%s $(CFLAGS) $(INCLUDE)\n\n",
            objlist[i], srclist[i]);
  }
}

void write_multi_rules(FILE * fp, MC2_VARS* o)
{
  char** mainlist = slist_new();
  _objlist2mainlist(&mainlist, o->pobjlist);
  
  char** srclist = *o->psrclist;
  char** objlist = *o->pobjlist;
  // copy mains from objlist without .o
  
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## Rules Section\n");
  fprintf(fp, "\n");
  fprintf(fp, "all: $(MAIN) $(HDR) $(SRC)\n");
  fprintf(fp, "\n");
  
  for(int i = 0; i < slist_count(objlist); i++)
  {
    fprintf(fp, "$(BINDIR)/%s: $(OBJDIR)/%s.o\n", mainlist[i], mainlist[i]);
    fprintf(fp, "\t@echo\n");
    fprintf(fp, "\t@echo \"Linking %s\"\n", mainlist[i]);
    fprintf(fp, "\t$(LINK) $(BINDIR)/%s $(OBJDIR)/%s.o $(LDFLAGS) $(LIB)\n", mainlist[i], mainlist[i]);
    fprintf(fp, "\t$(POST)\n");
    fprintf(fp, "\n");
    fprintf(fp, "$(OBJDIR)/%s.o: $(SRCDIR)/%s $(HDR)\n", mainlist[i],
            srclist[i]);
    fprintf(fp, "\t@echo\n");
    fprintf(fp, "\t@echo \"Compiling %s\"\n", mainlist[i]);
    fprintf(fp,
            "\t$(COMPILE) $(OBJDIR)/%s.o $(SRCDIR)/%s $(CFLAGS) $(INCLUDE)\n\n",
            mainlist[i], srclist[i]);
  }
  slist_delete(&mainlist);
}

void write_main_additional_targets(FILE * fp)
{
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## Additional Targets\n");
  fprintf(fp, "\n");
  fprintf(fp, "update:\n");
  fprintf(fp, "\t@mc2 -update\n\n");
  fprintf(fp, 
          "# example targets\n"
              "#mc2-semiclean:\n"
              "#\t@rm -f *~ */*~ */*/*~\n"
              "\n"
              "#mc2-clean:\n"
              "#\t@rm -f $(MAIN)\n"
              "#\t@rm -f $(OBJ)\n"
              "#\t@rm -f *~ */*~ */*/*~ */*/*/*~\n"
         );
}

void write_main_userdef(FILE * fp)
{
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## User Defined Targets\n\n");

  // if no foot defined, use defaults
  if(!getenv("MK_FOOT"))
  {
    fprintf(fp, "clean:\n");
    fprintf(fp, "\t@rm -f $(MAIN)\n");
    fprintf(fp, "\t@rm -f $(OBJ)\n");
    fprintf(fp, "\t@rm -f *~ */*~ */*/*~ */*/*/*~\n\n");
  }
}

void write_project_targets(FILE* fp, MC2_VARS* o)
{

  fprint_hline(fp, '#', 64);
  fprintf(fp, "## Additional Targets\n\n");
  fprintf(fp, 
          "# example targets\n"
              "#mc2-semiclean:\n"
              "#  @SUBDIRS make semiclean\n"
              "\n"
              "#mc2-clean:\n"
              "#  @SUBDIRS make clean\n\n"
         );

  fprintf(fp, "update:\n");
  fprintf(fp, "\tmc2 -update\n");
  fprintf(fp, "\tmake update-recursive\n\n");
  fprintf(fp, "update-recursive:\n");
  for(int i = 0; i < slist_getCount(o->psubdirs); i++)
    fprintf(fp, "\tcd %s && make update\n", (*o->psubdirs)[i]);
  fprintf(fp, "\n");
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## User Defined\n\n");
}


void write_mk_foot(FILE * fp)
{
  if(getenv("MK_FOOT"))
    fprintf(fp, getenv("MK_FOOT"));
}

int write_makefile(MC2_VARS* o)
{
  FILE *fp = fopen("Makefile", "r");
  if(fp)
  {
    system("cp -f Makefile Makefile~");
    fclose(fp);
  }
  fp = fopen("Makefile", "w");
  if(!fp)
  {
    fprintf(stderr, "Can't create Makefile\n");
    return 1;
  }
  // for project and normal
  write_makehdr(fp);
  if(!streq(o->outname, "PROJECT")) 
  {
    // is a main, not a project makefile
    write_main_mod_section(fp);
    write_mk_head(fp);
    
    write_assign_main_outname(fp, o);    
    write_assign_main_dirs(fp, o);
    write_assign_main_compiler(fp, o);
    write_assign_main_linker(fp, o);
    
    // handle MULTI as may be appropriate
    if(streq("MULTI", o->outname))
    {
      write_multi_list_section(fp, o);
      write_multi_rules(fp, o);
    }
    else // normal
    {
      write_main_list_section(fp, o);
      write_main_rules(fp, o);
    }    
//    write_main_mod_section(fp);    
    write_main_additional_targets(fp);
    write_main_userdef(fp);
    
    write_mk_foot(fp);
    write_divider(fp);
  }
  else
  {
    // is a project makefile
    write_mk_head(fp);
    
    write_project_file_lists(fp, o); // define PROJECT_FILES and MAKEFILES    
    write_project_rules(fp, o);
    write_project_targets(fp, o);
    
    write_mk_foot(fp);
  }
  fclose(fp);
  return 0;
}

int write_mc2_def(MC2_VARS* o)
{
  FILE *fp = fopen("mc2.def", "r");
  if(fp)
  {
    system("cp -f mc2.def mc2.def~");
    fclose(fp);
  }
  fp = fopen("mc2.def", "w");
  if(!fp)
  {
    fprintf(stderr, "Can't create mc2.def\n");
    return 1;
  }
  if(!streq(o->outname, "PROJECT")) 
  {
    write_assign_main_outname(fp, o);    
    write_assign_main_dirs(fp, o);
    write_assign_main_compiler(fp, o);
    write_assign_main_linker(fp, o);
    write_main_userdef(fp);
  }
  else
  {
    write_assign_main_outname(fp, o);    
  }
  fclose(fp);
  return 0;
}


// return env var if set else default
const char *check_override(const char *varname, const char *default_val)
{
  char *env = getenv(varname);
  if(env != 0)
    return env;
  else
    return default_val;
}

static void delete_str(char *buf, const char *s)
{
  char *p = strstr(buf, s);
  int len = strlen(s);
  if(p)
    strcpy(p, p + len);
}

// splits string at comment char if present so the switches at the
// end of the compiler command come before the comment.
static void do_end_sw(char* var, const char* endswitch)
{
  char sw[16];
  char tmp1[256];
  char tmp2[256];
  char* p;

  sprintf(sw, " %s", endswitch);
  strcpy(tmp1, var);
  
  // cut tmp1 at # and copy second half to tmp2 removing
  // preceeding space from tmp1 if present and always adding
  // the space back in tmp2.
  p = strchr(tmp1, '#');
  if(p)
  {
    *p = 0;
    sprintf(tmp2, " #%s", p+1);    
    p = tmp1 + strlen(tmp1) - 1;
    if(*p == ' ')
      *p = 0;    
  }
  else
    *tmp2 = 0;
  
  // remove the switch from wherever it is and restore it at the end
  delete_str(tmp1, sw);
  strcat(tmp1, sw);
  
  // join the cut string back together 
  sprintf(var, "%s%s", tmp1, tmp2);
}

// remove and restore endswitches in the right positions. TODO: Probably should 
// issue warnings if these need to be added or reordered.
void check_end_switch(char *var, const char *endswitch1, const char *endswitch2)
{
  
  if(endswitch1)
    do_end_sw(var, endswitch2);
  if(endswitch2)
    do_end_sw(var, endswitch2);
}


// write the rules per subdir for make, starting with all: tag
void write_project_rules(FILE* fp, MC2_VARS* o)
{
  fprint_hline(fp, '#', 64);
  fprintf(fp, "## Rules section\n");
  fprintf(fp, "\n");
  fprintf(fp, "all: $(PROJECT_FILES)\n");
  for(int i = 0; i < slist_getCount(o->psubdirs); i++)
    fprintf(fp, "\tcd %s && make\n", (*o->psubdirs)[i]);
  fprintf(fp, "\n");
  
  for(int i = 0; i < slist_getCount(o->psubdirs); i++)
  {
    char* s = (*o->psubdirs)[i];
    fprintf(fp, "%s/Makefile: %s/mc2.def\n", s, s);
    fprintf(fp, "\tcd %s && mc2 -update\n\n", s);
  }
}
   

// creates missing dirs and loads file lists
void setup_main(MC2_VARS* o)
{
  int err;
  err = create_dirs(o->srcdir, o->objdir, o->bindir);
      
    // load the files lists from the directories
  err = read_srclist(o->psrclist, o->srcdir);
  if(err)
    doerr("Can't read SRC files");
      
  err = read_hdrlist(o->phdrlist, o->srcdir);
  if(err)
    doerr("Can't read HDR files");
      
  err = create_objlist(o->pobjlist, o->psrclist);
  if(err) doerr("Can't create OBJ list");
}

// writes a header for a NONE type mc2.def file
void write_NONE_type()
{
#include "type_NONE_str.dat"
  FILE* fp = fopen("mc2.def", "w");
  if(!fp)
    return;
  write_makehdr(fp);
  fprintf(fp, type_NONE_str);
  fclose(fp);
}
