#include <sys/stat.h>
#include <ctype.h>    // isalpha() ..
#include <stdlib.h>   // getenv(), canonicalize_file_name(),...
#include <LQ/filename.h>
#include <LQ/tmpstr.h>
#include <LQ/slist.h>
#define NO_PARSER_WARN 1
#include <LQ/Parser.h> // include last in list to avoid problems with global 'obj' def
#include "utils.h"

#include "config.h"

const char *appname;
const char *pwd;    // initial dir

// true if mc2.def file is in current directory
bool
have_mc2_def ()
{
  return filename_exists ("mc2.def");
}

// uses shell command
int create_dirs (const char *srcdir, const char *objdir, const char *bindir)
{
  int err;
  char buf[256];
  sprintf (buf, "mkdir -p %s", srcdir);
  err = system (buf);
  sprintf (buf, "mkdir -p %s", objdir);
  err |= system (buf);
  sprintf (buf, "mkdir -p %s", bindir);
  err |= system (buf);
  return err;
}

// uses 'find' to get the file names.
int read_srclist (char ***psrclist, const char *srcdir)
{
  int err = 0;
  
  if (chdir (srcdir))
    return 1;
  // find src -type f -name *.c -o -name *.cxx -o -name *.cpp -o -name *.cc
    char syscmd[256];
    sprintf (syscmd,
             "find * -maxdepth 0 -type f -name '*.c' -o -name '*.cxx' -o -name '*.cpp' -o -name '*.cc' 2>/dev/null");
    err = slist_pipeRead (psrclist, syscmd);
    chdir (pwd);
    return err;
}

int read_hdrlist (char ***phdrlist, const char *srcdir)
{
  int err = 0;
  if (chdir (srcdir))
    return 1;
  // find src -type f -name *.c -o -name *.cxx -o -name *.cpp 
    char syscmd[256];
    sprintf (syscmd,
             "find * -maxdepth 0 -type f -name '*.h' -o -name '*.hxx' -o -name '*.hpp' 2>/dev/null");
    err = slist_pipeRead (phdrlist, syscmd);
    chdir (pwd);
    return err;
}

int create_objlist (char ***pobjlist, char ***psrclist)
{
  int i;      // loop counter
  char tmp[256];    // temp string
  char *p;      // scrath pointer
  char *filename;
  for (i = 0; i < slist_count (*psrclist); i++)
  {
    filename = (*psrclist)[i];
    // change ext from .* to .o
    strcpy (tmp, filename);
    p = strrchr (tmp, '.');
    strcpy (p, ".o");
    
    // tmp now = objdir/filename.o
    slist_append (pobjlist, tmp);
  }
  return 0;
}

// find mc2.def files in dirs one level down for recursive 'make'
int create_subdirlist(char*** psubdirs)
{
  
  char syscmd[256];
  int err = 0;
  sprintf (syscmd,
           "find * -maxdepth 1 -mindepth 1 -type f -name mc2.def 2>/dev/null");
  err = slist_pipeRead (psubdirs, syscmd);
  if(err)
    slist_clear(psubdirs);
  else
  {
    char* p;
    char* s;
    for(int i = 0; i < slist_getCount(psubdirs); i++)
    {
      s = (*psubdirs)[i];
      // remove the file name and just keep the dir name
      p = strrchr(s, '/');
      if(p)
        *p = 0;
      else
        slist_remove(psubdirs, i);
    }
  }
  return err;
}


/**
 *  Returns true if filename \a s matches pattern \a p. The following
 *  glob syntax is used by pattern:
 * 
 *  - "\x" quotes the character x so it must be matched exactly
 *  - "*" matches any sequence of 0 or more characters.
 *  - "?" matches any single character.
 * [Based on code by Rich Salz]
 */

bool dosex_glob(const char *s, const char *p)
{
  int nesting;
  
  for (;;)
  {
    switch (*p++)
    {
      case '?':   // match any single character
      {
        if (!*s++)
          return false;
        break;
      }   
      case '*':   // match 0-n of any characters
      {
        if (!*p)
          return true;  // do trailing * quickly
        while (!dosex_glob(s, p))
          if (!*s++)
            return false;
        return true;
      }     
      ///      case '|':   // skip rest of |pattern|pattern} when called recursively
      case ',':
      {
        for (nesting = 0; *p && nesting >= 0;)
        {
          switch (*p++)
          {
            case '\\':
              if (*p)
                p++;
              break;
            case '{':
              nesting++;
              break;
            case '}':
              nesting--;
              break;
          }
        }
        break;
      }
      case 0:   // end of pattern
        return !*s;
   
      case '\\':    // quote next character
        if (*p)
          p++;
      default:
        if (*s++ != *(p - 1))
          return false;
        break;
    }
  }
}

/*! Returns true if pattern(s) match.  Each pattern is parsed 
 * according to the rules for dosex_glob().
 */
bool dosex(const char *name, const char *patterns)
{
  char tmp[256];
  strcpy (tmp, patterns);
  int len = strlen (tmp);
  
  // remove spaces before or after commas
  for (int i = 0; i < len; i++)
  {
    if (tmp[i] == ' ' && tmp[i + 1] == ',')
    {
      memmove (tmp + i, tmp + i + 1, len - i);
      len--;
      i--;
    }
    else if (tmp[i] == ',' && tmp[i + 1] == ' ')
    {
      memmove (tmp + i + 1, tmp + i + 2, len - i - 1);
      len--;
      i--;
    }
  }
  const char *seps = "\";,";
  char *p = strtok (tmp, seps);
  for (; p != 0; p = strtok (0, seps))
  {
    if (dosex_glob(name, p))
      return true;
  }
  return false;
}

void doerr(const char *msg)
{
  fprintf(stderr, "%s\n", msg);
  exit(1);
}

const int mc2_dir_max = 256;
char mc2_dir[mc2_dir_max];

/// sets absolute path to the executable in buf, returns 0 on success
int setExecDir(char* buf, int buflen)
{
  *buf = 0;
  int tmplen = 1023;
  int execlen = 0;
  int pid = getpid();
  char tmp[tmplen + 1];
  
  sprintf(tmp, "/proc/%d/exe", pid);  
  execlen = readlink(tmp, tmp, tmplen);
  
  if(execlen >= tmplen)
    return 1; // too long
  
  // terminate for strrchr() and trim dirname (minus file name)
  // and get new execlen.
  tmp[execlen] = 0; 
  char* p = strrchr(tmp, '/');
  if(p) *p = 0;
  execlen = strlen(tmp);
  tmp[execlen] = 0;
  
  
  if((execlen+1) >= buflen)
    return 1; // can't fit in user's buffer
  
  strcpy(buf, tmp);
  return 0;
}

// Prints character ch n times, to stdout or any other
// FILE*
void fprint_hline(FILE* fp, char ch, int n)
{
  for(int i = 0; i < n; i++)
    fprintf(fp, "%c", ch);
  fprintf(fp, "\n");
}


// Takes a quoted comma separated list of buttons (minus spaces),
// a message and possible extended xmessage commands.  Prints the
// message in a centered window and returns the label of the button 
// clicked as a recycling tmpstr() string that must be used before 
// 32 new ones are created.
char* msgbox(const char* buttons, const char* msg, const char* ext)
{
  // There doesn't seem to be an easy way to get the return codes back 
  // from xmessage without a shell, so we'll use '-print' to figure out 
  // which button was pressed and capture the output in an slist.
  // Note that '-default <button>' may return an empty string, i.e., ""
  // as will stuff like alt-f4.  This is just a toy but possibly useful.
  
  char** pipestr = slist_new();
  char* cmd = tmpstr(strlen(msg) + 256, "");
  char* res;
  sprintf(cmd, "xmessage %s -center -print -buttons \"%s\" \"%s\"", ext == 0 ? "" : ext, buttons, msg);
  slist_pipeRead(&pipestr, cmd);
  slist_append(&pipestr, "");           // easier than checking if list is empty
  res = tmpstr(32, "%s", pipestr[0]);   // only want first string
  slist_delete(&pipestr);
  return res;
}

// returns true if target exists in file, assumes targstr is
// EXACTLY how it appears in a makefile, including the colon,
// with or without spaces.
int find_target(const char* targstr, const char* filename)
{
  int res = 0;
  char** mkfile = slist_new();
  slist_fileRead(&mkfile, filename);
  if(slist_count(mkfile))
  {
    for(int i = 0; i < slist_count(mkfile); i++)
    {
      char* p = slist_get(&mkfile, i);
      if(strstr(p, targstr) == p)
      {
        res++;
        break;
      }
    }
  }  
  slist_delete(&mkfile);
  return res;
}


void tst()
{
  // for testing parts of the build
}